@@ -45,6 +45,10 @@
 
 struct TgaReader
 {
+    Q_DISABLE_COPY(TgaReader)
+
+    TgaReader() = default;
+
     virtual ~TgaReader() {}
     virtual QRgb operator()(QIODevice *s) const = 0;
 };
@@ -52,13 +56,13 @@ struct TgaReader
 struct Tga16Reader : public TgaReader
 {
     ~Tga16Reader() {}
-    QRgb operator()(QIODevice *s) const
+    QRgb operator()(QIODevice *s) const override
     {
         char ch1, ch2;
         if (s->getChar(&ch1) && s->getChar(&ch2)) {
             quint16 d = (int(ch1) & 0xFF) | ((int(ch2) & 0xFF) << 8);
             QRgb result = (d & 0x8000) ? 0xFF000000 : 0x00000000;
-            result |= (d & 0x7C00 << 6) | (d & 0x03E0 << 3) | (d & 0x001F);
+            result |= ((d & 0x7C00) << 6) | ((d & 0x03E0) << 3) | (d & 0x001F);
             return result;
         } else {
             return 0;
@@ -68,7 +72,7 @@ struct Tga16Reader : public TgaReader
 
 struct Tga24Reader : public TgaReader
 {
-    QRgb operator()(QIODevice *s) const
+    QRgb operator()(QIODevice *s) const override
     {
         char r, g, b;
         if (s->getChar(&b) && s->getChar(&g) && s->getChar(&r))
@@ -80,7 +84,7 @@ struct Tga24Reader : public TgaReader
 
 struct Tga32Reader : public TgaReader
 {
-    QRgb operator()(QIODevice *s) const
+    QRgb operator()(QIODevice *s) const override
     {
         char r, g, b, a;
         if (s->getChar(&b) && s->getChar(&g) && s->getChar(&r) && s->getChar(&a))
@@ -144,9 +148,7 @@ QTgaFile::QTgaFile(QIODevice *device)
         mErrorMessage = tr("Seek file/device for image read failed");
         return;
     }
-    int bytes = device->read((char*)mHeader, HeaderSize);
-    if (bytes != HeaderSize)
-    {
+    if (device->read(reinterpret_cast<char*>(mHeader), HeaderSize) != HeaderSize) {
         mErrorMessage = tr("Image header read failed");
         return;
     }
@@ -170,9 +172,7 @@ QTgaFile::QTgaFile(QIODevice *device)
         return;
     }
     char footer[FooterSize];
-    bytes = mDevice->read((char*)footer, FooterSize);
-    if (bytes != FooterSize)
-    {
+    if (mDevice->read(reinterpret_cast<char*>(footer), FooterSize) != FooterSize) {
         mErrorMessage = tr("Could not read footer");
     }
     if (qstrncmp(&footer[SignatureOffset], "TRUEVISION-XFILE", 16) != 0)
