@@ -51,17 +51,22 @@ class QTiffHandler : public QImageIOHandler
 public:
     QTiffHandler();
 
-    bool canRead() const Q_DECL_OVERRIDE;
-    bool read(QImage *image) Q_DECL_OVERRIDE;
-    bool write(const QImage &image) Q_DECL_OVERRIDE;
+    bool canRead() const override;
+    bool read(QImage *image) override;
+    bool write(const QImage &image) override;
 
-    QByteArray name() const Q_DECL_OVERRIDE;
+    QByteArray name() const override;
 
     static bool canRead(QIODevice *device);
 
-    QVariant option(ImageOption option) const Q_DECL_OVERRIDE;
-    void setOption(ImageOption option, const QVariant &value) Q_DECL_OVERRIDE;
-    bool supportsOption(ImageOption option) const Q_DECL_OVERRIDE;
+    QVariant option(ImageOption option) const override;
+    void setOption(ImageOption option, const QVariant &value) override;
+    bool supportsOption(ImageOption option) const override;
+
+    bool jumpToNextImage() Q_DECL_OVERRIDE;
+    bool jumpToImage(int imageNumber) Q_DECL_OVERRIDE;
+    int imageCount() const Q_DECL_OVERRIDE;
+    int currentImageNumber() const Q_DECL_OVERRIDE;
 
     enum Compression {
         NoCompression = 0,
@@ -70,6 +75,7 @@ public:
 private:
     void convert32BitOrder(void *buffer, int width);
     const QScopedPointer<QTiffHandlerPrivate> d;
+    bool ensureHaveDirectoryCount() const;
 };
 
 QT_END_NAMESPACE
